use {
    crate::{error::Error, B64Alphabet},
    std::io::{ErrorKind, Read, Write},
};

pub struct Decoder<R: Read, W: Write> {
    reader: R,
    writer: W,
    alphabet: B64Alphabet,
}

impl<R: Read, W: Write> Decoder<R, W> {
    pub fn new(reader: R, writer: W, alphabet: Option<B64Alphabet>) -> Self {
        Self {
            reader,
            writer,
            alphabet: alphabet.unwrap_or_default(),
        }
    }

    /// Decodes the bytes provided by `self.reader` and writes the resulting
    /// bytes into `self.writer`.
    /// # Errors
    /// - io failure will return `Error::Io`
    /// - a character not in the alphabet, or encountering any character after
    ///   the first occurance of the padding character, will return `Error::IllegalChar`
    /// - if the reader comes up short of the final four byte block will return
    ///   `Error::MissingPadding`
    /// - `intError` should never be returned. If this error is recieved the code
    ///   is somehow broken. Please file a bug.
    pub fn decode(mut self) -> Result<W, Error> {
        let mut byte_reader = self.reader.bytes();
        loop {
            let mut in_buf = [0_u8; 4];
            let mut out_buf = [0_u8; 3];
            let mut num = 0_u32;
            let mut n_bytes = 0;
            while n_bytes < 4 {
                if let Some(byte) = byte_reader.next() {
                    match byte {
                        Ok(b) => {
                            if b == b'\n' {
                                continue;
                            }
                            in_buf[n_bytes] = b;
                            n_bytes += 1;
                        }
                        Err(e) => match e.kind() {
                            ErrorKind::UnexpectedEof => break,
                            ErrorKind::Interrupted => continue,
                            _ => return Err(e.into()),
                        }
                    }
                } else {
                    break;
                }
            }
            match n_bytes {
                0 => break,
                4 => {}
                _ => return Err(Error::MissingPadding),
            }
            let mut bytes = 0;
            for (i, &c) in in_buf.iter().enumerate() {
                let c = c.into();
                num <<= 6;
                if c == self.alphabet.pad() {
                    continue;
                }
                if i != bytes {
                    return Err(Error::IllegalChar(c));
                }
                let Some(idx) = self.alphabet.get(c) else {
                    return Err(Error::IllegalChar(c));
                };
                num |= u32::try_from(idx)?;
                bytes += 1;
            }
            let olen = bytes * 6 / 8;
            for i in (0..3).rev() {
                out_buf[i] = (num & 0xff) as u8;
                num >>= 8;
            }
            self.writer.write_all(&out_buf[0..olen])?;
        }
        Ok(self.writer)
    }
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn decode() {
        let mut output = vec![];
        let mut decoder = Decoder::new("SGVsbG8sIFdvcmxk".as_bytes(), output, None);
        output = decoder.decode().unwrap();
        assert_eq!(String::from_utf8(output.clone()).unwrap(), "Hello, World");
        output.clear();
        decoder = Decoder::new("SGVsbG8sIFdvcmxkIQ==".as_bytes(), output, None);
        output = decoder.decode().unwrap();
        assert_eq!(String::from_utf8(output.clone()).unwrap(), "Hello, World!");
        output.clear();
        decoder = Decoder::new("SGVsbG8sIFdvcmxkIQo=".as_bytes(), output, None);
        output = decoder.decode().unwrap();
        assert_eq!(
            String::from_utf8(output).unwrap(),
            "Hello, World!\n"
        );
    }
}
